import pony
import SCons.Builder
from bag import MongoConnectionInfo

def __build_connection_info(env):
    connection_info = MongoConnectionInfo() 
    if 'mongo_db' in env:    
        connection_info.host = env['mongo_db']
    if 'mongo_port'in env:     
        connection_info.port = env['mongo_port']
    if 'mongo_user' in env:
        connection_info.user = env['mongo_user']
    if 'mongo_port' in env:
        connection_info.pwd = env['mongo_pwd']
        connection_info.controlled_access  = True

    return connection_info

def __charge(target, source, env):
     
    pony.charge_json_file(str(source[0]), __build_connection_info(env))
    return None

def __deliver(target, source, env):
    
    fname=str(source[0])
    with open(fname, 'r') as f:
        pony.deliver_json(f.read(), __build_connection_info(env))

    return None

charge_artifact = SCons.Builder.Builder( action = __charge,
                                        src_suffix = '.json')

deliver_dependencies = SCons.Builder.Builder(  action = __deliver,
                                             src_suffix = '.json')                                        

def establish_contact (scons_environment):
    scons_environment['BUILDERS']['charge'] = charge_artifact
    scons_environment['BUILDERS']['deliver'] = deliver_dependencies
    return scons_environment


class Requirement(object):
    def __init__(self, json_decoded):
        self.invalid = False
        try:
            del json_decoded['DEPENDENCIES']
        except KeyError:
            pass

        self.__dict__ = json_decoded

    def __eq__(self, other):
        return self.__hash__() == other.__hash__()

    def __lt__(self, other):
        return str(self) < str(other)

    def __hash__(self):
        return hash(str(self))

    def __repr__(self):
        return str(self.NAME)

    def __str__(self):
        text_val = "{ "
        meta_request = self.meta_request()
        keylist = list(meta_request.keys())
        keylist.sort()

        text_val += str(keylist[0]) + ":" + str(meta_request[keylist[0]])

        for k in keylist[1:]:
            text_val += ", " + str(k) + ":" + str(meta_request[k])

        text_val += " }"

        return text_val


    def perfect_match(self, other):
        for x in [prop for prop in self.__dict__ if prop in other.__dict__]:
            if self.__dict__[x] != other.__dict__[x]:
                print ("unmatched " + x)
                return False
        return True

    def alternative(self, other):
        return self.NAME == other.NAME

    def meta_request(self):
        stripped = dict (self.__dict__)
        try:
            del stripped['TEST']
        except KeyError:
            pass

        try:
            del stripped['_id']
        except KeyError:
            pass

        try:
            del stripped['parent']
        except KeyError:
            pass

        try:
            del stripped['package']
        except KeyError:
            pass

        return stripped


def make_direct( meta_request, parent=None ):
    requirements = [Requirement(r) for r  in meta_request]
    if parent:
        for r in requirements:
            r.parent = parent

    return requirements

def dot_graph(requirements):
    graph = "\ndigraph {\n"
    for r in [r for r in requirements if not hasattr(r, "parent")]:
        graph += "  \"" + repr(r) + "\"\n"

    graph += "\n"

    for r in [r for r in requirements if hasattr(r, "parent")]:
        graph += "  \"" + repr(r) + "\" -> \"" + repr(r.parent) + "\"\n"

    graph += "}"
    return graph



import unittest
import collections
import dependencies as dep
import bag 
from bag import Bag, to_dot_string
import networkx as nx
import time
from operator import xor


class TestRequirementFunctionalities (unittest.TestCase):
  def setUp(self):
    pass

  def tearDown(self):
    pass

  def test_can_construct_requirement_using_dict (self):
    """
    """
    
    req = dep.Requirement ( { 
                              "NAME" : "DEP-1",
                              "VERSION" : "1.1.1"
                            }
    )
    
    self.assertTrue( hasattr(req, "NAME") )
    self.assertEqual( req.NAME, "DEP-1")
    
    self.assertTrue( hasattr(req, "VERSION") )
    self.assertEqual( req.VERSION,  "1.1.1") 
    
    
  def test_requirements_equal_on_lose_match(self):
    req1 = dep.Requirement ( { 
                              "NAME" : "DEP-1",
                              "VERSION" : "1.0.1"
                            }
    )
    
    req2 = dep.Requirement ( { 
                              "NAME" : "DEP-1",
                              "VERSION" : "1.1.1"
                            }
    )
    
    req_match_1 = dep.Requirement ( { 
                              "NAME" : "DEP-1",
                              "VERSION" : "1.0.1"
                            }
    )
    
    self.assertTrue(req1.alternative(req2))
    self.assertFalse(req1.perfect_match(req2))
    self.assertEqual(req1, req_match_1)
    self.assertTrue(req1, req_match_1)
    
    
  def test_direct_dependencies_extraction(self):
    """
    pony can receive a metainformation object and build direct requirements.
    The metainformations object can be constructed by decoding a json string into a python object the default way.
    """
    direct_dependencies = dep.make_direct(  [
                                              {
                                                "NAME" : "DEP-1",
                                                "VERSION" : "1.1.1"
                                              },
                                              {
                                                "NAME" : "DEP-2",
                                                "VERSION" : "1.2.3"
                                              }
                                            ]
    )
    
    # construct the collection of expected objects to check for match
    expected_objects = [
                        dep.Requirement( {
                                        "NAME" : "DEP-1",
                                        "VERSION" : "1.1.1"
                        }),
                                      
                        dep.Requirement( {
                                        "NAME" : "DEP-1",
                                        "VERSION" : "1.1.1"
                        })
                        ]
    
    # direct_dependencies is an iterable object
    self.assertTrue( isinstance(direct_dependencies, collections.Iterable) )
    # it contains exactly two elements
    self.assertEqual( len(direct_dependencies), 2)
    # it contains all the expected objects
    for expected in expected_objects:
      self.assertTrue(expected in direct_dependencies) 
  
  def test_can_represent_dependency_graph_using_dot(self):  
    r1 = dep.Requirement( {"NAME" : "DEP-1", "VERSION" : "1.1.1"} )   
    r2 = dep.Requirement( { "NAME" : "DEP-2", "VERSION" : "1.0.1" })
    
    r2_1= dep.Requirement( { "NAME" : "DEP-2-1", "VERSION" : "1.0.1" })
    r2_1.parent = r2
    r2_2= dep.Requirement( { "NAME" : "DEP-2-2", "VERSION" : "1.0.6" })
    r2_2.parent = r2
    
    
    r2_1_1= dep.Requirement( { "NAME" : "DEP-2-1-1", "VERSION" : "1.0.6" })
    r2_1_1.parent = r2_1
    
    dot_graph = dep.dot_graph([r1, r2, r2_1, r2_2, r2_1_1])
    self.assertEqual(dot_graph, """
digraph {
  "DEP-1"
  "DEP-2"

  "DEP-2-1" -> "DEP-2"
  "DEP-2-2" -> "DEP-2"
  "DEP-2-1-1" -> "DEP-2-1"
}""")
    print("DOT:  "+ dot_graph)
      

class TestBagRequirementDiscover (unittest.TestCase):
  def setUp(self):
    self.connected_bag = Bag()
    self.connected_bag.silent = True;
    repo = self.connected_bag
    repo.collection.delete_many({"TEST" : ""})
    self.dep_1 = {"NAME" : "DEP-1", "VERSION" : "0.1.0", "TEST" : ""}
    self.dep_2 = {"NAME" : "DEP-2", "VERSION" : "0.1.0", "TEST" : "", 
                      "DEPENDENCIES" : [ 
                                        {"NAME" : "DEP-2-1", "VERSION" : "1.1.1"},
                                        {"NAME" : "DEP-2-2", "VERSION" : "1.2.3"}
                                       ] }
    repo.charge(None, self.dep_1 )
    repo.charge(None, {"NAME" : "DEP-2-1", "VERSION" : "1.1.1", "TEST" : ""} )
    repo.charge(None, {"NAME" : "DEP-2-2", "VERSION" : "1.2.3", "TEST" : ""} )
    repo.charge(None,  self.dep_2 )    
    repo.charge(None, {"NAME" : "P1", "VERSION" : "1.0.0", "TEST" : "", 
                      "DEPENDENCIES" : [ 
                                        {"NAME" : "DEP-1", "VERSION" : "0.1.0"},
                                        {"NAME" : "DEP-2", "VERSION" : "0.1.0"}
                                       ] } )

    repo.charge(None, {"NAME" : "P0", "VERSION" : "1.0.0", "TEST" : "", 
                      "DEPENDENCIES" : [ 
                                        {"NAME" : "DEP-1", "VERSION" : "0.1.0"},
                                        {"NAME" : "DEP-2-1", "VERSION" : "1.1.1", "TEST" : ""}
                                       ] } )
    
    self.connected_bag.silent = False;

  def tearDown(self):
    self.connected_bag.collection.delete_many({"TEST" : ""})

  def test_bag_query_for_direct_project_dependencies(self):
    """
    Check the bag can discover direct dependencies for a given requirement
    """
    requirements = self.connected_bag.requirements_for (dep.Requirement ({ "NAME" : "P1", "VERSION" : "1.0.0"}))
    
    self.assertEqual( len(requirements), 2);
    self.assertTrue( dep.Requirement(self.dep_1) in requirements, msg= str(dep.Requirement(self.dep_1) ) + "\nRESULTS: " + str(requirements) )
    self.assertTrue( dep.Requirement(self.dep_2) in requirements, msg= str(dep.Requirement(self.dep_2) ) + "\nRESULTS: " + str(requirements) )


  def test_bag_query_for_direct_project_dependencies_only_direct_dep(self):
    """
    Check the bag can discover direct dependencies for a given requirement
    """
    requirements, gr = self.connected_bag.requirements_discover (dep.Requirement ({ "NAME" : "P0", "VERSION" : "1.0.0"}))
    self.assertEqual(
      len(requirements),
      gr.number_of_nodes(),
      msg='number of dependency graph node do not match detected requirements number.'
    )
    self.assertEqual( len(requirements), 3, msg='found requirements: ' + str(requirements));
    self.assertTrue( dep.Requirement(self.dep_1) in requirements, msg= str(dep.Requirement(self.dep_1) ) + "\nRESULTS: " + str(requirements) )

  def test_bag_query_for_project_dependencies(self):
    
    requirements, gr = self.connected_bag.requirements_discover(dep.Requirement ({ "NAME" : "P1", "VERSION" : "1.0.0"}))
    self.assertEqual(
      len(requirements),
      gr.number_of_nodes(),
      msg='number of dependency graph node do not match detected requirements number.'
    )
    self.assertEqual( len(requirements), 5, msg="discovered requirements: " + str(requirements));
    self.assertTrue( dep.Requirement(self.dep_1) in requirements)
    self.assertTrue( dep.Requirement(self.dep_2) in requirements)
    
    dep_2_1 = [r for r in requirements if r.NAME == "DEP-2-1"][0]
    self.assertEqual( dep_2_1.parent, dep.Requirement(self.dep_2) )
    
  def test_bag_query_for_multiple_dependencies(self):
    """
    Check the bag can discover the dependencies graph for a list of projects.
    This is the case of building the graph of dependencies of a project
    while it is still not mantained by pony. Need this to find out the dependency graph given the
    set of dependencies listed into the json metainformations fils.
    """
    requirements, gr = self.connected_bag.requirements_discover (
      [
        dep.Requirement ( { "NAME" : "P0", "VERSION" : "1.0.0"}),
        dep.Requirement ( { "NAME" : "DEP-2"})
      ]
    )

    self.assertEqual(
      len(requirements),
      gr.number_of_nodes(),
      msg='number of dependency graph node do not match detected requirements number.\n' + str(requirements) + '\nvs\n'+ str(gr)
    )
    self.assertEqual( len(requirements), 5, msg='found requirements: ' + str(requirements));
    self.assertTrue( dep.Requirement(self.dep_1) in requirements, msg= str(dep.Requirement(self.dep_1) ) + "\nRESULTS: " + str(requirements) )


  def test_bag_query_for_project_cyrcular_dependency(self):
    """
    Checks the bag can deal, *using simple policy* with circular dependencies. Basically it do not loop infinitelly at least.
    """
    
    self.dep_circular = { "NAME" : "DEP_CIR", "VERSION" : "0.0.3", "TEST" : "",
                          "DEPENDENCIES" :  [
                                              { 
                                                "NAME" : "DEP-2",
                                                "VERSION" : "0.1.0"
                                              }
                                            ]
                        }
                        
    self.prj =          { "NAME" : "P", "VERSION" : "0.0.3", "TEST" : "",
                          "DEPENDENCIES" :  [
                                              { 
                                                "NAME" : "P1",
                                                "VERSION" : "1.0.0"
                                              },
                                              { 
                                                "NAME" : "DEP_CIR",
                                                "VERSION" : "0.0.3"
                                              }
                                            ]
                        }
                        
                        
                        
    self.connected_bag.charge(None, self.dep_circular)
    self.connected_bag.charge(None, self.prj)
    
    requirements, gr = self.connected_bag.requirements_discover( dep.Requirement ({ "NAME" : "P", "VERSION" : "0.0.3"})) 
    self.assertEqual(len(requirements), 7, msg="discovered requirements: " + str(requirements) ) 
    
    graph = self.connected_bag.requirements_graph(dep.Requirement( {"NAME" : "P", "VERSION" : "0.0.3"}))
    print ("check this")
    print(bag.to_dot_string(graph))

    
class TestCanDealWithAlternativeDependencies(unittest.TestCase):
  """
  Tests the pony project requirement detection mechanism in case of multiple alternative project dependency. 
  The pony shall be able to:
    - constrain dependency choise in order to satisfy different (but compatible in some way) request from multiple dependant projects.
    - arbitrary choose a dependency project in case multiple project instancies can fulfill requests of dependant.
    - provide feedback on the dependency graph constructed with indication of project chosed and the one considered invalid due to constraint on requirements. 
    - of course pony also provide feedback in case It is not possible to satisfy requirements due to existing constraints.
  """
  def setUp(self):
    """ Setup a reference scenario. 

    +----+              +----+                              +--------------+
    |    |------------->| PA |----------------------------->| DEP-1 v0.2.0 |
    |    |              +----+                          --->|              |
    |    |               |                             /    +--------------+
    |    |               |                            /
    | PB |               |       +--------------+    /      +--------------+
    |    |               +------>| DEP-2 v0.1.0 |---------->|              |
    |    |                       +--------------+           | DEP-1 v0.1.0 |
    |    |                                                  |              |
    |    |------------------------------------------------->|              | 
    |    |                                                  +--------------+
    +----+ 

    Notice that the dependency of PB from DEP-1 v0.1 and PA conflicts with the dependency of PA from DEP-1 v0.2.

    """

    
    self.connected_bag = Bag()
    self.connected_bag.silent = True;
    repo = self.connected_bag

    self.connected_bag.collection.delete_many({"TEST" : ""})
    self.dep_1 =    {"NAME" : "DEP-1", "VERSION" : "0.1.0", "TEST" : "" }
    self.dep_1_v2 = {"NAME" : "DEP-1", "VERSION" : "0.2.0", "TEST" : ""}
    self.dep_2 =    {"NAME" : "DEP-2", "VERSION" : "0.1.0", "TEST" : "",
                      "DEPENDENCIES" : [ 
                                        {"NAME" : "DEP-1", "VERSION" : {"$gte" : "0.0.1"}  }
                                      ] }
                                      
    self.prj =      {"NAME" : "PA", "VERSION" : "0.1.0", "TEST" : "",
                      "DEPENDENCIES" : [ 
                                        {"NAME" : "DEP-1", "VERSION" : "0.2.0"},
                                        {"NAME" : "DEP-2", "VERSION" : "0.1.0"}
                                      ] }
    
    self.pb =       {   "NAME" : "PB", "VERSION" : "0.1.0", "TEST" : "",
                      "DEPENDENCIES" : [ 
                                        {"NAME" : "PA", "VERSION" : "0.1.0"},
                                        {"NAME" : "DEP-1", "VERSION" : "0.1.0"}
                                      ] }
    
    self.connected_bag.charge(None, self.dep_1)
    self.connected_bag.charge(None, self.dep_1_v2)
    self.connected_bag.charge(None, self.dep_2)
    self.connected_bag.charge(None, self.prj)
    self.connected_bag.charge(None, self.pb)
    

  def tearDown(self):
    self.connected_bag.collection.delete_many({"TEST" : ""})
    pass


  def test_only_one_alternative_dependency(self):
    """
    Test pony can chose from alternative dependencies when it is constrained by some other requirement
    """
    
    requirements, gr = self.connected_bag.requirements_discover(dep.Requirement( {"NAME" : "PA"} ))
    self.assertEqual(len(requirements), 3, msg = """
      expected three requirement (including project itself) but found """ + str(len(requirements)) + """
      They actually are: """ + str(requirements) + """
      Requirements graph is:
      """ + to_dot_string(gr) )
    
    req_d1_v2 = dep.Requirement(self.dep_1_v2)

    self.assertTrue( req_d1_v2 in requirements, msg="""RESULTS: 
    """+ str(requirements) + """
      Requirements graph is:
      """ + to_dot_string(gr) )
  
  def test_multiple_alternative_resolved_random(self):
    """
    pony can resolve multiple alternative for dependencies using any of the available dependant package.
    """
    
    requirements, gr = self.connected_bag.requirements_discover(dep.Requirement( {"NAME" : "DEP-2"}))
    self.assertEqual(len(requirements), 2, msg = """
      expected two requirements (including project itself) but found """ + str(len(requirements)) + """
      They actually are: """ + str(requirements) )
      
    req_d1 = dep.Requirement(self.dep_1)
    req_d1_v2 = dep.Requirement(self.dep_1_v2)
    self.assertTrue( xor(req_d1 in requirements, req_d1_v2 in requirements), msg="""Actual requirements:
    """+ str(requirements) )

  def test_multiple_alternative_direct_requirements(self):
    """
    pony can detect and resolve multiple alternative direct requests using any of the available dependant package.
    The test require DEP-1 project witch is in the bag in versioni 0.1.0 and 0.2.0 so the request must be
    translated into a graph witch describe the request for both and then the graph reduced pruning alternatives.
    """
    
    requirements, gr = self.connected_bag.requirements_discover(dep.Requirement( {"NAME" : "DEP-1"}))
    self.assertEqual(len(requirements), 1, msg = """
      expected only one requirements (including project itself) but found """ + str(len(requirements)) + """
      They actually are: """ + str(requirements) )

    self.assertEqual(len(gr.nodes()), 2, msg= "expected two alternative version of DEP-1 detected into the bag.\n"+ to_dot_string(gr))
      
    req_d1 = dep.Requirement(self.dep_1)
    req_d1_v2 = dep.Requirement(self.dep_1_v2)
    self.assertTrue( xor(req_d1 in requirements, req_d1_v2 in requirements), msg="""Actual requirements:
    """+ str(requirements) )
    
  def test_conflicting_transitive_dependencies(self):
    """When pony detect conflicting transitive dependencies it must raise exception. 
    The excetion raised must carry the annotated dependency graph for log and debugging purpose. Here we check the dot representation of the dependency graph.
    It may change in future, it is up to anyone who will make changes to check the updated version of this graph is descriptive enough to show the conflicts discovered.
    """
    self.maxDiff=None
    with self.assertRaises(bag.ImpossibleConfigurationException):
        requirements, gr = self.connected_bag.requirements_discover(dep.Requirement( {"NAME" : "PB"}))
    try:
        self.connected_bag.requirements_discover(dep.Requirement( {"NAME" : "PB"}))
        self.assert_("expect it to raise exception, didn't it failed yet?\n"+ bag.to_dot_string())
    except bag.ImpossibleConfigurationException as exception_raised:
        expected_dot_graph = """digraph {
    "{ NAME:DEP-1, VERSION:0.1.0 }" [penwidth=3 ];
    "{ NAME:DEP-1, VERSION:0.2.0 }" [penwidth=3 ];
    "{ NAME:DEP-2, VERSION:0.1.0 }" [penwidth=3 ];
    "{ NAME:PA, VERSION:0.1.0 }" [penwidth=3 ];
    "{ NAME:PB, VERSION:0.1.0 }" [penwidth=3 ];

    "{ NAME:DEP-2, VERSION:0.1.0 }" -> "{ NAME:DEP-1, VERSION:0.1.0 }"[color=black penwidth=3];
    "{ NAME:DEP-2, VERSION:0.1.0 }" -> "{ NAME:DEP-1, VERSION:0.2.0 }"[color=black penwidth=3];
    "{ NAME:PA, VERSION:0.1.0 }" -> "{ NAME:DEP-1, VERSION:0.2.0 }"[color=black penwidth=3];
    "{ NAME:PA, VERSION:0.1.0 }" -> "{ NAME:DEP-2, VERSION:0.1.0 }"[color=black penwidth=3];
    "{ NAME:PB, VERSION:0.1.0 }" -> "{ NAME:DEP-1, VERSION:0.1.0 }"[color=black penwidth=3];
    "{ NAME:PB, VERSION:0.1.0 }" -> "{ NAME:PA, VERSION:0.1.0 }"[color=black penwidth=3];
}
"""
        self.assertEqual( expected_dot_graph, to_dot_string(exception_raised.gr), msg= "Unmatched graph representation, check changes please. It is expected to have conflicting dependencies requests." )
        
if __name__ == "__main__":
  unittest.main()


#!python3

import unittest
import shutil
import os


import pymongo
import pony
from bag import to_dot_string


def ensure_path(path):
    """ensure the full path received exists"""
    try:
        os.makedirs(path)
    except FileExistsError:
        pass

def touch(file_path):
    """create an empty file at the specified path"""
    tempf = open(file_path, 'w')
    tempf.close()


def touch_some_files():
    """Create an input folder structure"""
    print('setting up')
    ensure_path('test/temp-data/subfolder/subsubfolder')
    touch('test/temp-data/one.txt')
    touch('test/temp-data/two.txt')
    touch('test/temp-data/subfolder/ignore.tx')
    touch('test/temp-data/subfolder/three.txt')
    touch('test/temp-data/subfolder/subsubfolder/four.txt')

def erase_temp_files():
    """Destroy temporary folder structure"""
    shutil.rmtree('test/temp-data')

class TestMongoStoreOp(unittest.TestCase):
    """Test the store operation on real mongo db"""

    def setUp(self):
        touch_some_files()

    def tearDown(self):
        """Destroy temporary folder structure"""
        erase_temp_files()

        connection = pymongo.MongoClient('localhost', 27017)
        db = connection.pony_store
        collection = db.packages

        collection.delete_many({"name" : "temp"})


    def test_save_package_and_metadata_in_mongo(self):
        """
        Checks both 'raw' package data and related metadata are correctly stored in mongo db.
        This also checks you cannot store multiple time projects with same metadata.
        """
        # establish mongo connection and retrieve the pony_store collection
        connection = pymongo.MongoClient('localhost', 27017)
        collection = connection.pony_store.packages

        # prepare some packaging instructions and metadata
        instructions = [
            pony.pack(
                'test/temp-data',
                'test-data',
                '*.txt'
            ),
            pony.pack(
                'test/temp-data/subfolder/subsubfolder',
                'moved-folder',
                '*.txt'
            )
        ]

        some_metadata = {"name" : "temp", "version" : "0.1.3", "target" : "x86"}

        # clean the collection from any document with same (test) metadata
        collection.delete_many(some_metadata)
        pony.charge(instructions, some_metadata)

        # checks mongo has a correspondant document stored!
        package = collection.find({"name" : "temp", "version" : "0.1.3", "target" : "x86"})
        self.assertTrue(package[0]['package'])

        # and that it prevent you from store multiple artifact with same metadata!
        with self.assertRaises(pony.bag.YetInBag):
            pony.charge(instructions, some_metadata)

    def test_save_package_and_complex_metadata(self):
        """
        Checks both 'raw' package data and related metadata are correctly stored in mongo db.
        (more complex metadata: complex dependency specified).
        This also checks you cannot store multiple time projects with same metadata but different
        dependencies.
        """
        # establish mongo connection and retrieve the pony_store collection
        connection = pymongo.MongoClient('localhost', 27017)
        collection = connection.pony_store.packages

        # prepare some packaging instructions and metadata
        instructions = [
            pony.pack(
                'test/temp-data',
                'test-data',
                '*.txt'
            ),
            pony.pack(
                'test/temp-data/subfolder/subsubfolder',
                'moved-folder',
                '*.txt'
            )
        ]

        some_metadata = {
            "name" : "temp",
            "version" : "0.1.4",
            "target" : "x86",
            "DEPENDENCIES" : [
                {
                    "NAME" : "A_PROJECT",
                    "VERSION" : {"$gte" : "0.0.1"}
                }
            ]
        }

        # clean the collection from any document with same (test) metadata
        collection.delete_many(some_metadata)
        pony.charge(instructions, some_metadata)

        # checks mongo has a correspondant document stored!
        package = collection.find({"name" : "temp", "version" : "0.1.4", "target" : "x86"})
        self.assertTrue(package[0]['package'])

        # and that it prevent you from store multiple artifact with same metadata!
        with self.assertRaises(pony.bag.YetInBag):
            pony.charge(instructions, some_metadata)

    def test_insert_based_on_json(self):
        """
        Checks pony charge a box completelly described by mean of json string and by
        boxing instructions referring to files on the filesyste.
        """
        # establish mongo connection and retrieve the pony_store collection
        connection = pymongo.MongoClient('localhost', 27017)
        collection = connection.pony_store.packages

        pony.charge(json="""
        {
            "name"    : "temp", 
            "version" : "0.1.3", 
            "target"  : "x86",
            
           "__BOX_INSTRUCTIONS__" : [
                                    {
                                        "FROM"      : "test/temp-data", 
                                        "TO"        : "test-data", 
                                        "FILTER"    : "*.txt"
                                    },
                                    {
                                        "FROM"      : "test/temp-data/subfolder/subsubfolder", 
                                        "TO"        : "moved-folder", 
                                        "FILTER"    : "*.txt"
                                    }
            ]
        }
        """)

        # checks mongo has a correspondant document stored!
        package = collection.find({"name" : "temp", "version" : "0.1.3", "target" : "x86"})
        self.assertTrue(package[0]['package'])

    def test_insert_based_on_json_file(self):
        """
        Checks pony charge a box completelly described by mean of json file and by
        boxing instructions referring to files on the filesystem.
        """
        # establish mongo connection and retrieve the pony_store collection
        connection = pymongo.MongoClient('localhost', 27017)
        collection = connection.pony_store.packages

        json_content = """
        {
            "name" : "temp", 
            "version" : "0.1.3", 
            "target" : "x86",
            
           "__BOX_INSTRUCTIONS__" : [
                                    {
                                        "FROM"      : "test/temp-data", 
                                        "TO"        : "test-data", 
                                        "FILTER"    : "*.txt"
                                    },
                                    {
                                        "FROM"      : "test/temp-data/subfolder/subsubfolder", 
                                        "TO"        : "moved-folder", 
                                        "FILTER"    : "*.txt"
                                    }
            ]
        }
        """
        with open("temp.json", "w") as out_file:
            out_file.write(json_content)

        pony.charge(file="temp.json")
        # checks mongo has a correspondant document stored!
        package = collection.find({"name" : "temp", "version" : "0.1.3", "target" : "x86"})
        self.assertTrue(package[0]['package'])

class TestMongoMetarequestQuery(unittest.TestCase):
    def setUp(self):
        pony.silent = True
        touch_some_files()
        self.connection = pymongo.MongoClient('localhost', 27017)
        self.collection = self.connection.pony_store.packages

        instructions = [
            pony.pack('test/temp-data', 'test-data', '*.txt'),
            pony.pack('test/temp-data/subfolder/subsubfolder', 'moved-folder', '*.txt')
        ]

        self.collection.delete_many(
            {"NAME" : "dep-v1"}
        )
        self.collection.delete_many(
            {"NAME" : "prj61"}
        )

        meta = {"NAME" : "dep-v1",
                "VERSION" : "1.0.0"}
        pony.charge(instructions, meta)

        meta = {"NAME" : "dep-v1",
                "VERSION" : "0.1.0"}
        pony.charge(instructions, meta)

        meta = {"NAME" : "prj61",
                "VERSION" : "1.3.0"}

        pony.charge(instructions, meta)

        meta = {"NAME" : "dep-v1",
                "VERSION" : "1.1.0"}
        pony.charge(instructions, meta)

    def tearDown(self):
        pony.silent = True
        erase_temp_files()
        #also erase files created as results
        try:
            shutil.rmtree('test/results-data')
        except FileNotFoundError:
            pass
        self.collection.delete_many(
            {"NAME" : "dep-v1"}
        )
        self.collection.delete_many(
            {"NAME" : "prj61"}
        )
        self.connection.close()

    def test_deliver_op_version_exact(self):
        """
        Test a ssuccess case, it must successfully
        unbox a stored package with request metadata
        """

        instructions = [
            pony.unpack('test-data', 'test/results-data'),
            pony.unpack('moved-folder', 'test/results-data')
        ]

        meta_request = {"NAME" : "dep-v1", "VERSION" :  "1.0.0"}
        pony.deliver(instructions, meta_request)

        self.assertTrue(
            os.path.exists('test/results-data/one.txt')
        )
        self.assertTrue(
            os.path.exists('test/results-data/two.txt')
        )
        self.assertTrue(
            os.path.exists('test/results-data/subfolder/three.txt')
        )

    def test_deliver_op_return_created_files(self):
        """
        Test a ssuccess case, it must successfully unbox a stored package
        with request metadata
        """

        instructions = [
            pony.unpack('test-data', 'test/results-data'),
            pony.unpack('moved-folder', 'test/results-data')
        ]
        meta_request = {"NAME" : "prj61", "VERSION" :  "1.3.0"}
        created_files, dep_graph = pony.deliver(instructions, meta_request)

        for fname in ['test/results-data/one.txt', 'test/results-data/two.txt', 'test/results-data/subfolder/three.txt']:
            self.assertTrue(
                fname in created_files,
                msg = str(fname) + ' not in '+ str(created_files) + '\nDependency graph is\n' + to_dot_string(dep_graph)
            )

    def test_deliver_op_version_greater(self):
        """
        Test a success case, it must successfully unbox a stored package
        with request metadata (version greater then)
        """

        instructions = [
            pony.unpack('test-data', 'test/results-data'),
            pony.unpack('moved-folder', 'test/results-data')
        ]

        meta_request = {"NAME" : "dep-v1", "VERSION" : {"$gte" : "0.9.0"}}
        pony.deliver(instructions, meta_request)

        self.assertTrue(
            os.path.exists('test/results-data/one.txt')
        )
        self.assertTrue(
            os.path.exists('test/results-data/two.txt')
        )
        self.assertTrue(
            os.path.exists('test/results-data/subfolder/three.txt')
        )

    def test_deliver_op_fail_version_exact(self):
        """
        Test a fail case, it must raise NotInBag exception because
        there isn't a stored package with request metadata
        """

        instructions = [
            pony.unpack('test-data', 'test/results-data'),
            pony.unpack('moved-folder', 'test/results-data')
        ]

        meta_request = {"NAME" : "dep-v1", "VERSION" :  "1.0.1"}
        with self.assertRaises(pony.bag.NotInBag):
            pony.deliver(instructions, meta_request)

    def test_deliver_op_version_greater_from_json(self):
        """
        Checks pony can deliver boxes based on json a requirement
        description (actually just a single requirement description)
        and the instructions for the unboxing.
        In this case the requireiment description is complex, it is like: version greater then...
        """
        import json
        json_string = """
        [
                {"NAME" : "dep-v1", "VERSION" :  {"$gte" : "0.9.0"} }
        ]"""
        meta_request = json.loads(json_string)

        instructions = [
            pony.unpack('test-data', 'test/results-data'),
            pony.unpack('moved-folder', 'test/results-data')
        ]

        requirements, gr = pony.deliver(instructions, meta_request[0])
        instructions = [
            pony.pack(
                'test/temp-data',
                'test-data',
                '*.txt'
            ),
            pony.pack(
                'test/temp-data/subfolder/subsubfolder',
                'moved-folder',
                '*.txt'
            )
        ]
        self.assertTrue(
            os.path.exists('test/results-data/one.txt')
        )
        self.assertTrue(
            os.path.exists('test/results-data/two.txt')
        )
        self.assertTrue(
            os.path.exists('test/results-data/subfolder/three.txt')
        )

        # checks the log from the deliver operation. This will check the all compatible versions of dependencies are taken into account
        # and that only one is seleceted to resolve dependency. 

        self.assertEqual("""digraph {
    "{ NAME:dep-v1, VERSION:1.0.0 }" [penwidth=3 color=blue ];
    "{ NAME:dep-v1, VERSION:1.1.0 }" [penwidth=3 ];

}
""" , to_dot_string(gr), msg=to_dot_string(gr))


    def test_deliver_op_version_greater_from_simple_and_complete(self):
        """
        Checks pony can deliver boxes based on json description of project metadata:
            - metadata strictly concerning the project
            - dependency metadata
            - unbox instructions
        In this case the requreiment description is complex, it is like: version greater then...
        """

        import json
        json_string = """
        {
            "NAME" : "TEST",
            "DEPENDENCIES" :
                            [
                                {"NAME" : "dep-v1", "VERSION" :  {"$gte" : "0.9.0"} }

                            ],

            "__UNBOX_INSTRUCTIONS__" : [
                 {
                    "FROM" : "test-data", 
                    "TO"   :  "test/results-data",
                    "FILTER" : "*"
                 },
                 {
                    "FROM" :  "moved-folder", 
                    "TO"   :  "test/results-data",
                    "FILTER" : "*"
                 }
            ]

        }"""

        pony.deliver_json(json_string)

        self.assertTrue(
            os.path.exists('test/results-data/one.txt')
        )
        self.assertTrue(
            os.path.exists('test/results-data/two.txt')
        )
        self.assertTrue(
            os.path.exists('test/results-data/subfolder/three.txt')
        )

    def test_deliver_op_multiple_dependencies(self):
        """
        Checks can retrieve multiple dependencies using only one set of unpack instructions.

        Checks pony can deliver boxes based on json description of project dependencies metadata:
            - dependency metadata
            - unbox instructions

        In this case the requreiment description is complex, it is like: version greater then...
        In this case there are multiple acceptable dependency project version. Pony shall
        chose any of those and success.

        Checks also that the requirement graph constructed from json dependency description, is the same as the one contructed
        based on lower level dependency description (by means of python objects). 

        
        """

        import json
        json_string = """
        [
                {   "NAME" :  "dep-v1",
                    "VERSION" :  {"$gte" : "0.9.0"} 
                },

                {   "NAME"      : "prj61",
                    "VERSION"   : "1.3.0"
                }
        ]"""

        # construct the requirement description object from json and from python object.
        meta_request = json.loads(json_string)
        meta_request_raw = [{"NAME" : "dep-v1", "VERSION" : {"$gte" : "0.9.0"}}, {"NAME" : "prj61", "VERSION" : "1.3.0"}] 

        instructions = [
            pony.unpack('test-data', 'test/results-data'),
            pony.unpack('moved-folder', 'test/results-data')
        ]

        # ask pony to deliver dependencies as described from the requirements constructed from json and from raw python object.
        files, gr = pony.deliver(instructions, meta_request)
        files_raw, gr_raw = pony.deliver(instructions, meta_request_raw)

        fail_message = "The requirement graph constructed from json string is:\n" + to_dot_string(gr) + "\nwhile the one constructed from equivalent python object is:\n" + to_dot_string(gr_raw)
        self.assertEqual( to_dot_string(gr), to_dot_string(gr_raw), msg= fail_message )

        self.assertTrue(
            os.path.exists('test/results-data/one.txt')
        )
        self.assertTrue(
            os.path.exists('test/results-data/two.txt')
        )
        self.assertTrue(
            os.path.exists('test/results-data/subfolder/three.txt')
        )

    def test_pony_report_chosed_dependencies_as_networkx_graph(self):
        """
        The deliver operation return as a result the list of created
        files as a result of the dependency resolution process
        It also must produce a log of the dependency resolution process
        in the form of a networkx graph.
        """

        import json
        encoded_string = """
        {
            "NAME" : "TEST",
            "DEPENDENCIES" :
                            [
                                {"NAME" : "dep-v1", "VERSION" :  {"$gte" : "0.9.0"} }

                            ],

            "__UNBOX_INSTRUCTIONS__" : [
                 {
                    "FROM" : "test-data", 
                    "TO"   :  "test/results-data",
                    "FILTER" : "*"
                 },
                 {
                    "FROM" :  "moved-folder", 
                    "TO"   :  "test/results-data",
                    "FILTER" : "*"
                 }
            ]

        }"""

        file_list, gr = pony.deliver_json(encoded_string)

        self.assertEqual("""digraph {
    "{ NAME:dep-v1, VERSION:1.0.0 }" [penwidth=3 color=blue ];
    "{ NAME:dep-v1, VERSION:1.1.0 }" [penwidth=3 ];

}
""" , to_dot_string(gr), msg=to_dot_string(gr))


if __name__ == '__main__':
    unittest.main()
    


import unittest
import cli
import bag 
from test.facilities import *
import shutil

class TestStringMethods(unittest.TestCase):

    def setUp(self):
        """
        Setup multiple (but minimal) projects and their dependencies.

        +-----+         +-----+             +-----+         +-----+
        | Av0 | ------->| Bv1 | ----------->| Cv1 | ------->| Dv1 |
        +-----+         +-----+             +-----+         +-----+
           |                                   |
           |            +-----+                |         +-----+
           |            |     |                +-------->|     |
           +----------->| Bv2 |------------------------->| Dv2 |
           |            |     |---------------+  +------>|     |
           |            +-----+               |  |       +-----+
           |                                  v  |
           |            +-----+             +-------+       +-----+
           +----------->| Bv3 | ----------->| Cv2   |------>| Dv3 |
                        +-----+             +-------+       +-----+
        
        
        Each of these projects contains only one file named test_<project-name>_v<version>
        and only two real metadata NAME and VERSION plus a tag to erase them  from the server once 
        the tests are done (TEST).
        """

        mongo_clean()
        projects =  {"A": [0], "B": [1,2,3], "C":[1,2], "D": [1,2,3] }
        metainfo = {"A": {}, "B": {}, "C":{}, "D": {} }
# prepare metainformations for Av0
        metainfo["A"][0] = """
{
    "NAME" : "A",
    "VERSION" : "0",
    "TEST" : "",

    "DEPENDENCIES" : [
        {
            "NAME" : "B",
            "VERSION" : { "$lte": "3",  "$gte" : "1" }
        }
    ],
    
    "__BOX_INSTRUCTIONS__" : [
                        {
                            "FROM" 		: "", 
                            "TO" 		: "package-folder", 
                            "FILTER" 	: "*"
                        }
    ],

    "__UNBOX_INSTRUCTIONS__" : [
                        {
                            "FROM" 		: "package-folder", 
                            "TO" 		: "result-folder", 
                            "FILTER" 	: "*"
                        }
    ]

}""" 

        metainfo["B"][1] = """
{
    "NAME" : "B",
    "VERSION" : "1",
    "TEST": "",

    "DEPENDENCIES" : [
        {
            "NAME" : "C",
            "VERSION" : "1" 
        } 
    ],
    
    "__BOX_INSTRUCTIONS__" : [
                        {
                            "FROM" 		: "", 
                            "TO" 		: "package-folder", 
                            "FILTER"   : "*.file" 
                        }
    ],

    "__UNBOX_INSTRUCTIONS__" : [
                        {
                            "FROM" 		: "package-folder", 
                            "TO" 		: "result-folder", 
                            "FILTER"   : "*.file"
                        }
    ]

}"""


        metainfo["B"][2] = """
{
    "NAME" : "B",
    "VERSION" : "2",
    "TEST": "",

    "DEPENDENCIES" : [
        {
            "NAME" : "D",
            "VERSION" : "2" 
        },
        
        {
            "NAME" : "C",
            "VERSION" : "2" 
        } 
    ],
    
    "__BOX_INSTRUCTIONS__" : [
                        {
                            "FROM" 		: "", 
                            "TO" 		: "package-folder", 
                            "FILTER"   : "*.file"   
                        }
    ],

    "__UNBOX_INSTRUCTIONS__" : [
                        {
                            "FROM" 		: "package-folder", 
                            "TO" 		: "result-folder", 
                            "FILTER"   : "*.file"   
                        }
    ]

}"""

        metainfo["B"][3] = """
{
    "NAME" : "B",
    "VERSION" : "3",
    "TEST": "",

    "DEPENDENCIES" : [
        {
            "NAME" : "C",
            "VERSION" : "2" 
        } 
    ],
    
    "__BOX_INSTRUCTIONS__" : [
                        {
                            "FROM" 		: "", 
                            "TO" 		: "package-folder", 
                            "FILTER"   : "*.file"   
                        }
    ],

    "__UNBOX_INSTRUCTIONS__" : [
                        {
                            "FROM" 		: "package-folder", 
                            "TO" 		: "result-folder", 
                            "FILTER"   : "*.file"   
                        }
    ]

}"""

        metainfo["C"][1] = """
{
    "NAME" : "C",
    "VERSION" : "1",
    "TEST": "",

    "DEPENDENCIES" : [
        {
            "NAME" : "D",
            "VERSION" : { "$lte": "2",  "$gte" : "1" } 
        }
    ],
    
    "__BOX_INSTRUCTIONS__" : [
                        {
                            "FROM" 		: "", 
                            "TO" 		: "package-folder", 
                            "FILTER"   : "*.file"   
                        }
    ],

    "__UNBOX_INSTRUCTIONS__" : [
                        {
                            "FROM" 		: "package-folder", 
                            "TO" 		: "result-folder", 
                            "FILTER"   : "*.file"   
                        }
    ]

}"""

        metainfo["C"][2] = """
{
    "NAME" : "C",
    "VERSION" : "2",
    "TEST": "",

    "DEPENDENCIES" : [
        {
            "NAME" : "D",
            "VERSION" : "3"
        }
    ],
    
    "__BOX_INSTRUCTIONS__" : [
                        {
                            "FROM" 		: "", 
                            "TO" 		: "package-folder", 
                            "FILTER"   : "*.file"   
                        }
    ],

    "__UNBOX_INSTRUCTIONS__" : [
                        {
                            "FROM" 		: "package-folder", 
                            "TO" 		: "result-folder", 
                            "FILTER"   : "*.file"   
                        }
    ]

}"""

        metainfo["D"][1] = """
{
    "NAME" : "D",
    "VERSION" : "1",
    "TEST": "",
    
    "__BOX_INSTRUCTIONS__" : [
                        {
                            "FROM" 		: "", 
                            "TO" 		: "package-folder", 
                            "FILTER"   : "*.file"   
                        }
    ],

    "__UNBOX_INSTRUCTIONS__" : [
                        {
                            "FROM" 		: "package-folder", 
                            "TO" 		: "result-folder", 
                            "FILTER"   : "*.file"   
                        }
    ]

}"""

        metainfo["D"][2] = """
{
    "NAME" : "D",
    "VERSION" : "2",
    "TEST": "",

    "DEPENDENCIES" : [
    ],
    
    "__BOX_INSTRUCTIONS__" : [
                        {
                            "FROM" 		: "", 
                            "TO" 		: "package-folder", 
                            "FILTER"   : "*.file"   
                        }
    ],

    "__UNBOX_INSTRUCTIONS__" : [
                        {
                            "FROM" 		: "package-folder", 
                            "TO" 		: "result-folder", 
                            "FILTER"   : "*.file"   
                        }
    ]

}"""

        metainfo["D"][3] = """
{
    "NAME" : "D",
    "VERSION" : "3",
    "TEST": "",
    
    "__BOX_INSTRUCTIONS__" : [
                        {
                            "FROM" 		: "", 
                            "TO" 		: "package-folder", 
                            "FILTER"   : "*.file"   
                        }
    ],

    "__UNBOX_INSTRUCTIONS__" : [
                        {
                            "FROM" 		: "package-folder", 
                            "TO" 		: "result-folder", 
                            "FILTER"   : "*.file"   
                        }
    ]

}"""

        for p in projects:
            versions = projects[p]
            for v in versions:
                folder = "test/test-cli/test-projects/" + p +"v" + str(v)
                project_file = folder + "/test_" + p +"_v" + str(v) + ".file"
                ensure_path(folder)
                touch(project_file)

                with open(folder + "/meta.json", "w") as f:
                    f.write(metainfo[p][v])
                    
                self._store_project_at(folder)

    def tearDown(self):
        """
        Destroy temporary folder structure and remove any mongo documents (pony box) 
        inserted temporary for test purpose
        """
        shutil.rmtree('test/test-cli')
        mongo_clean()

    def _store_project_at(self, folder):
        # set the current working directory to the project folder to mimic
        # the common use case
        restore_cwd = os.getcwd()

        os.chdir(folder)
        command_line = "pony charge".split(' ')
        arguments = cli.parser.parse_args(command_line[1:])
        arguments.func(arguments)

        # restore the original cwd
        os.chdir(restore_cwd)

    def test_cli_store_op(self):
        """
        Checks projects have been stored during the setup fase with the correct (basic) meta informations
        given into meta.json files
        """

        with MongoChecker() as ck:
            self.assertTrue(ck.contain({"NAME" : "A", "VERSION" : "0"}))
            self.assertTrue(ck.contain({"NAME" : "B", "VERSION" : "1"}))
            self.assertTrue(ck.contain({"NAME" : "B", "VERSION" : "2"}))
            self.assertTrue(ck.contain({"NAME" : "B", "VERSION" : "3"}))
            self.assertTrue(ck.contain({"NAME" : "C", "VERSION" : "1"}))
            self.assertTrue(ck.contain({"NAME" : "C", "VERSION" : "2"}))
            self.assertTrue(ck.contain({"NAME" : "D", "VERSION" : "1"}))
            self.assertTrue(ck.contain({"NAME" : "D", "VERSION" : "2"}))
            self.assertTrue(ck.contain({"NAME" : "D", "VERSION" : "3"}))

    def test_cli_deliver_op_no_dep(self):
        """
        Checks the deliver operation requested to build projects without
        extra dependencies (projects named D in the example).

        The only file added after a deliver operation on this kind of projects should be 
        .pony, the one that lists files from other projects
        """

        with WorkingDirectorySet("test/test-cli/test-projects/Dv1") as project_folder:
            command_line = "pony deliver".split(' ')
            arguments = cli.parser.parse_args(command_line[1:])
            arguments.func(arguments)

            prj_folder_files = files_in()
            prj_subfolders = subfolders_in()
            self.assertEqual(3, len(prj_folder_files), msg="more files then expected in project folder they are: " + str(prj_folder_files) )
            self.assertEqual(0, len(prj_subfolders), msg="created new folders after deliver operation they are: " + str(prj_subfolders) )

    def test_cli_deliver_op_simple_dep(self):
        """
        Checks the deliver operation required to build simple projects with only one
        dependency and no alternatives, successfully prepare the project environment
        Will use Cv2 witch depends only on Dv3 or Dv2 as an example project.

        This test also checks that dependencies are stored on mongo collection the 
        proper way during store operation
        """

        with WorkingDirectorySet("test/test-cli/test-projects/Cv2") as project_folder:
            command_line = "pony deliver".split(' ')
            arguments = cli.parser.parse_args(command_line[1:])
            arguments.func(arguments)

            prj_folders_files = files_in()
            prj_subfolders = subfolders_in()

            self.assertEqual(3, len(prj_folders_files), msg="Expect no new file other then .pony created in project folder. File list actually is: " + str(prj_folders_files) )
            self.assertEqual(1, len(prj_subfolders), msg="Expected result-folder created as project root folder subdirectory given the __UNBOX_INSTRUCTION__ into project metadata. Subfolders are: " + str(prj_subfolders) )
            self.assertEqual("result-folder", prj_subfolders[0], msg="Expect the only project subfolder created to be result-folder" )
            self.assertEqual(1, len(files_in('result-folder')), msg="Expect one new file created in result-folder.File list actually is: " + str('result-folder') ) 

    def test_cli_deliver_op_simple_transitive_dep(self):
        """
        Checks the deliver operation when required to build projects with transitive dependencies.

        Here I'll use Bv1 which depends on Cv1 and in turn on Dv1 or Dv2 alternativelly. This will check pony detect and resolve
        correctly indirect dependencies. Still no constraing is required apart of dependency tree arbitrary pruning on alternatives.
        """

        with WorkingDirectorySet("test/test-cli/test-projects/Bv1"):
            command_line = "pony deliver".split(' ')
            arguments = cli.parser.parse_args(command_line[1:])
            file_list, gr = arguments.func(arguments)

            prj_folders_files = files_in()
            prj_subfolders = subfolders_in()

            self.assertEqual(bag.to_dot_string(gr), """digraph {
    "{ NAME:C, VERSION:1 }" [penwidth=3 ];
    "{ NAME:D, VERSION:1 }" [penwidth=3 color=blue ];
    "{ NAME:D, VERSION:2 }" [penwidth=3 ];

    "{ NAME:C, VERSION:1 }" -> "{ NAME:D, VERSION:1 }"[color=black penwidth=3];
    "{ NAME:C, VERSION:1 }" -> "{ NAME:D, VERSION:2 }"[color=black penwidth=3];
}
"""
            )

            self.assertEqual(3, len(prj_folders_files), 
                                msg="Expect no new file other then .pony created in project folder. File list actually is: " + str(prj_folders_files) )
            self.assertEqual(1, len(prj_subfolders),
                                msg="Expected result-folder created as project root folder subdirectory given the __UNBOX_INSTRUCTION__ into project metadata. Subfolders are: " + str(prj_subfolders) )
            self.assertEqual("result-folder", prj_subfolders[0],
                                msg="Expect the only project subfolder created to be result-folder" )
            self.assertEqual(2, len(files_in(folder="result-folder")), 
                                msg="Expect exactly 2 files created into the dependency destination folder, one for each projects. File list actually is: " + str(os.listdir("result-folder")) )
             

    def test_cli_deliver_op_complex_transitive_dep(self):
        """
        Checks the deliver operation when required to build projects with complex transitive dependencies.

        Here pony is supposed to use constraints to resolve dependencies that can
        fulfill all transitive dependencies. 

        I'll use Av0 the dependency graph can be seen as documentation of setup method.
        """

        with WorkingDirectorySet("test/test-cli/test-projects/Av0"):
            command_line = "pony deliver".split(' ')
            arguments = cli.parser.parse_args(command_line[1:])
            try:
                file_list, gr = arguments.func(arguments)
            except  bag.TransitiveDependencyUnreachable as e:
                print ("EXCEPTION EXCEPTION")
                print (bag.to_dot_string(e.gr))
                raise e


            prj_folders_files = files_in()
            prj_subfolders = subfolders_in()
            self.maxDiff = None
            result_graph = bag.to_dot_string(gr)
            self.assertEqual(result_graph, """digraph {
    "{ NAME:B, VERSION:1 }" [penwidth=3 color=blue ];
    "{ NAME:B, VERSION:2 }" [penwidth=3 color=blue ];
    "{ NAME:B, VERSION:3 }" [penwidth=3 ];
    "{ NAME:C, VERSION:1 }" [penwidth=3 color=blue ];
    "{ NAME:C, VERSION:2 }" [penwidth=3 ];
    "{ NAME:D, VERSION:1 }" [penwidth=3 color=blue ];
    "{ NAME:D, VERSION:2 }" [penwidth=3 color=blue ];
    "{ NAME:D, VERSION:3 }" [penwidth=3 ];

    "{ NAME:B, VERSION:1 }" -> "{ NAME:C, VERSION:1 }"[color=black penwidth=3];
    "{ NAME:B, VERSION:2 }" -> "{ NAME:C, VERSION:2 }"[color=black penwidth=3];
    "{ NAME:B, VERSION:2 }" -> "{ NAME:D, VERSION:2 }"[color=black penwidth=3];
    "{ NAME:B, VERSION:3 }" -> "{ NAME:C, VERSION:2 }"[color=black penwidth=3];
    "{ NAME:C, VERSION:1 }" -> "{ NAME:D, VERSION:1 }"[color=black penwidth=3];
    "{ NAME:C, VERSION:1 }" -> "{ NAME:D, VERSION:2 }"[color=black penwidth=3];
    "{ NAME:C, VERSION:2 }" -> "{ NAME:D, VERSION:3 }"[color=black penwidth=3];
}
"""
            )

            self.assertEqual(3, len(prj_folders_files), 
                                msg="Expect no new file other then .pony created in project folder. File list actually is: " + str(prj_folders_files) )
            self.assertEqual(1, len(prj_subfolders),
                                msg="Expected result-folder created as project root folder subdirectory given the __UNBOX_INSTRUCTION__ into project metadata. Subfolders are: " + str(prj_subfolders) )
            self.assertEqual("result-folder", prj_subfolders[0],
                                msg="Expect the only project subfolder created to be result-folder" )
            self.assertEqual(3, len(files_in(folder="result-folder")), 
                                msg="Expect exactly 3 files created into the dependency destination folder, one for each projects it depends from. File list actually is: " + str(os.listdir("result-folder")) )
             

        



        





"""
This module defines the command line interface to pony.
It provide facilities to properly end to end test the whole application. 
"""

import pony
import bag 
from bag import Bag, MongoConnectionInfo

import json
import os, sys
import time
from argparse import ArgumentParser, FileType

#meve the following in a proper module!
    
def charge(args):
    metadata = args.meta.read()
    connection_info = MongoConnectionInfo(args)
    
    try:
        pony.charge(db_connection_info = connection_info, json=metadata)
    except bag.YetInBag as ex:
        jobject = json.loads(metadata)
        try:
            del jobject['package']
        except:
            pass
        
        try:
            del jobject['__UNBOX_INSTRUCTIONS__']
        except:
            pass
            
        try:
            del jobject['__BOX_INSTRUCTIONS__']
        except:
            pass
                
        print ("""
charge operation failed: there is a conflicting package in the bag. 
  Metadata are:

  """ + json.dumps(jobject, indent=4))

def deliver(args):
    connection_info = MongoConnectionInfo(args)
    created_files, gr = pony.deliver_json(args.meta.read(), db_connection_info=connection_info)
    with open('.pony', 'w') as f:
        for fname in created_files:
            f.write(fname+'\n')
    
    return created_files, gr

def clear(args):
    with open('.pony', 'r') as f:
        for fname in f:
            os.remove(fname[:-1])


# setup the parser and its subparsers
parser = ArgumentParser( prog='pony')
subparsers = parser.add_subparsers()

# meta-informations source
parser.add_argument('--meta', type=FileType('r'), default='meta.json')
# mongo host
parser.add_argument('--mongo', default='localhost')
# mongo port
parser.add_argument('--port', type=int, default=27017)
# user
parser.add_argument('--user', default='')
# password
parser.add_argument('--pwd', default='')

charge_command_parser = subparsers.add_parser('charge')
charge_command_parser.set_defaults(func=charge, dest='func')

deliver_command_parser = subparsers.add_parser('deliver')
deliver_command_parser.set_defaults(func=deliver)

deliver_command_parser = subparsers.add_parser('clear')
deliver_command_parser.set_defaults(func=clear)


import unittest 

import os
import shutil
import io

import pony
from test.facilities import ensure_path, touch



class TestStringMethods(unittest.TestCase):

    def setUp(self):
        """Create an input folder structure"""
        print ('setting up')
        ensure_path('test/temp-data/subfolder/subsubfolder')
        touch('test/temp-data/one.txt')
        touch('test/temp-data/two.txt')
        touch('test/temp-data/subfolder/ignore.tx')
        touch('test/temp-data/subfolder/three.txt')
        touch('test/temp-data/subfolder/subsubfolder/four.txt')

    def tearDown(self):
        """Destroy temporary folder structure"""
        shutil.rmtree('test/temp-data')
        shutil.rmtree('test/results-data')

    def test_file_exist_after_mem_compand(self):
        """
        Try to compress to an in-memory tarball storing test folder structure as a rerooted structure.
        then expand again in a different destination folder. 
        Apply a filter on archived files: take only .txt files. In fact skip ignore.tx"""
        memory = io.BytesIO()
        
        instruction = pony.pack('test/temp-data', 'test-data', '*.txt')
        pony.gz_package(package_instructions= [instruction], mem=memory)
        
        memory = io.BytesIO(memory.getvalue())
        instruction = pony.unpack('test-data', 'test/results-data')
        pony.gz_unbox(unbox_instructions= [instruction], mem = memory)
        
        
        self.assertTrue (os.path.exists('test/results-data/one.txt'))
        self.assertTrue (os.path.exists('test/results-data/two.txt'))
        self.assertTrue (os.path.exists('test/results-data/subfolder/three.txt'))
        self.assertTrue (os.path.exists('test/results-data/subfolder/subsubfolder/four.txt'))
        
        self.assertFalse(os.path.exists('test/results-data/subfolder/ignore.tx'))

    def test_complex_archiving_rerooting(self):
        """
        Tests the compand operation when source folder structure is reorganized in a complex way.
        Some subfolder into different level folders. """
        memory = io.BytesIO()
        
        instructions = [ 
                        pony.pack('test/temp-data', 'test-data', '*.txt'),
                        pony.pack('test/temp-data/subfolder/subsubfolder', 'moved-folder', '*.txt')
                      ]
                  
        pony.gz_package(package_instructions= instructions, mem=memory)
        
        memory = io.BytesIO(memory.getvalue())
        instructions = [
                        pony.unpack('test-data', 'test/results-data'),
                        pony.unpack('moved-folder', 'test/results-data')
                       ]
        pony.gz_unbox(unbox_instructions= instructions, mem = memory)
        
        
        self.assertTrue (os.path.exists('test/results-data/one.txt'))
        self.assertTrue (os.path.exists('test/results-data/two.txt'))
        self.assertTrue (os.path.exists('test/results-data/subfolder/three.txt'))
        self.assertTrue (os.path.exists('test/results-data/four.txt'))
        
        # it also has not extracted the rerooted elements into their original structure
        self.assertFalse (os.path.exists('test/results-data/subfolder/subsubfolder/four.txt'))
        self.assertFalse(os.path.exists('test/results-data/subfolder/ignore.tx'))
        
if __name__ == '__main__':
    unittest.main()


import collections
import json
from config_constrainer import config_generator, ConfigurationImpossible
from  pymongo import MongoClient
import dependencies as dep
import networkx as nx

import sys
from bson.binary import Binary


class ImpossibleConfigurationException(Exception):
    """Exception raised in case pony can't find an acceptable dependency configuration."""
    def __init__(self, gr):
        self.gr = gr

    def __str__ (self):
        return """Can't fulfill all dependencies in a compatible way.
The dependency graph is
""" + to_dot_string(self.gr)


def no_dependencies():
    return [], nx.DiGraph()

def quote(s, qm = '"'):
    return qm+ str(s) + qm

def to_dot_string(G):
    """
    Construct a string that represent the graph using dot languale.
    It assumes G is a directed graph.
    """

    dot_string = "digraph {\n"
    # first print out any node to be sure to not forget any isolated node
    sorted_nodes = G.nodes()
    sorted_nodes.sort()
    for n in sorted_nodes:
        n_dep = ''
        try:
            n_dep = G.node[n]['ndep']
        except:
            pass

        #dot_string += "    " + quote( str(n) +  '( ndep: ' + str(n_dep)+ ')') + " [penwidth=3 "
        dot_string += "    " + quote( str(n)) + " [penwidth=3 "
        if('invalid' in G.node[n]):
            dot_string += "color=red "
        elif G.node[n]['pruned']:
            dot_string += "color=blue "
        dot_string += "];\n"

    dot_string += "\n"

    # then write down any edge
    sorted_edges = G.edges(data=True)
    sorted_edges.sort()
    for edge in sorted_edges:
        dot_string += "    " +  quote(edge[0]) + " -> " + quote(edge[1]) 
        if 'valid' in edge[2]:
            if edge[2]['valid']:
                dot_string += '[color=black penwidth=3]'
            else:
                dot_string += '[color=red penwidth=3]'

        dot_string += ";\n"
    dot_string += "}\n"

    return dot_string


class MongoConnectionInfo:
    def __init__(self, pony_arguments = None):
        if pony_arguments and pony_arguments.user:
            self.host = pony_arguments.mongo
            self.port = pony_arguments.port
            self.user = pony_arguments.user
            self.pwd = pony_arguments.pwd
            self.controlled_access  = True
        else:
            self.host = 'localhost'
            self.port = 27017
            self.controlled_access  = False
            
            
    def connect(self):
        cli= MongoClient(host = self.host, port=self.port)
        print('connecting to: ' + self.host + ' : ' + str(self.port) + ' [controlled:' + str(self.controlled_access) + ']' )
        if self.controlled_access:
            cli.pony_store.authenticate(self.user, self.pwd)
        return cli

class YetInBag (Exception):
    
    def __init__ (self, meta_informations):
        self.explain = meta_informations
        
    def __str__ (self):
        return """
        Error trying to insert package into the bag. 
                It contains yet a package with same metadata. Perhaps you want to increase your version.
        
        Received metadata:
          """ + self.explain 
        
    def __str__ (self):
        return """
        Error trying to insert package into the bag. 
                It contains yet a package with same metadata. Perhaps you want to increase your version.
        
        Received metadata:
          """ + str(self.explain) 

class NotInBag (Exception):

    def __init__ (self, meta_informations):
        self.explain = meta_informations
        
    def __str__ (self):
        return """
        Error trying to take package from the bag. 
                It do not contains yet a package with compatible metadata.
        
        Received metadata:
          """ + str(self.explain) 

class Bag:
    def __init__(self, db_connection_info=MongoConnectionInfo()):
        # setup the mongo connection
        self.connection = db_connection_info.connect()
        self.collection = self.connection.pony_store.packages
        self.silent = False

    def encode_requirements(self, meta):
        if "DEPENDENCIES" in meta.keys():
            meta["DEPENDENCIES"] = json.dumps(meta["DEPENDENCIES"])
        return meta

    def decode_requirements(self, meta):
        query_requirements = []
        requirements = []
        if "DEPENDENCIES" in meta.keys():
            query_requirements = json.loads(meta["DEPENDENCIES"])
            for query in query_requirements:
                mongo_cursor = self.collection.find(query)
                for r in mongo_cursor:
                    requirements.append(r)

            meta['DEPENDENCIES'] = requirements
        return requirements

    def check(self, meta):
        same_meta_cursor = self.collection.find(meta)
        fail = same_meta_cursor.count() > 0
        if fail:
            pass#print(same_meta_cursor[0])
        return fail

    def charge (self,  package, meta):
        meta = self.encode_requirements(meta)
        if(self.check(meta)):
            raise YetInBag(meta)
        # add content to metadata. it will become a rich package!
        if sys.version_info < (3, 0):
            meta['package'] = Binary(package)
        else:
            meta['package'] = package
        id = self.collection.insert_one(meta)
        
        del meta['package']
        if not self.silent:
            print ('package archived')
            print (str(meta))
    
    def _select_all(self, meta_request):
        """Select all alternative package matching the request"""

        if type(meta_request) is dep.Requirement:
            meta_request = meta_request.meta_request()

        matching_packages = self.collection.find(meta_request)
        if matching_packages.count() == 0:
            raise NotInBag(meta_request)
        
        # return the first matching package
        return matching_packages
            
    def _select(self,meta_request):
        """Select an arbitrary alternative requested package from the set of matching one"""
        selection = []
        matching_packages = self._select_all(meta_request)
        
        
        for p in matching_packages:
            selection.append(p)
        
        for p in selection:
            p = self.decode_requirements(p)
                
        # return the first matching package
        return selection[0] 
        
    
    def take(self, meta_request):
        # take the package from charged document
        package = self._select(meta_request)['package']
        return package
        
    def translate_requirements(self, requests, select_any = True):
        """Translates the requirements into an explicit requirement given the matching boxed into actually in the bag. """
        
        if not isinstance(requests, list):
            requests = [requests]    

        meta_requests = [r.meta_request() for r in requests]

        requirements =[]
        for meta_request in meta_requests:
            for p in self._select_all(meta_request):
                requirements.append(dep.Requirement(p))
            
        return requirements



    def requirements_for (self, request, select_any = True):
        """
        Retrieve additional dependencies for a requirements_for
        """
        dependant_projects = set()
        meta_request = request.meta_request()
        package = self._select(meta_request) if select_any else self._select_all(meta_request)
        requirements = []
        try:
            requirements = dep.make_direct(package['DEPENDENCIES'], parent=request)
        except KeyError:
            # no dependencies for this package!
            pass
            
        return requirements
      
    def requirements_graph (self, direct_requirements, known_nodes = []):
        gr = nx.DiGraph()

        if not isinstance(direct_requirements, collections.Iterable):
            direct_requirements = [direct_requirements]

        while len(direct_requirements) > 0:
            req = direct_requirements.pop()
            more_requirements = self.requirements_for(req)
            #more_requirements.remove(req)

            local_gr = nx.DiGraph()
            for child_req in more_requirements:
                gr.add_edge(req, child_req, valid=True)
                gr.node[child_req]['pruned'] = False

            gr.add_node(req)
            gr.node[req]['pruned'] = False
            _gr = to_dot_string(gr)
            
            # get the graph describing the direct requirements requirements
            # then join it with the one describing direct requirements (local)
            sub_gr = self.requirements_graph(more_requirements)
            
            gr = nx.compose(gr, sub_gr)
            
        return gr
    
    def acceptable_configurations(self, gr):
        if len(gr.nodes()) == 0:
            return [gr]
        
        acceptable_config_generator = config_generator(gr)
        try:
            configurations = [config for config in acceptable_config_generator.constraint() ]
        except ConfigurationImpossible:
            raise ImpossibleConfigurationException(gr)

        return configurations

    def first_acceptable_configuration(self, gr):
        if len(gr.nodes()) == 0:
            return [gr]
        
        acceptable_config_generator = config_generator(gr)
        try:
            config = next(acceptable_config_generator.constraint())
        except ConfigurationImpossible:
            raise ImpossibleConfigurationException(gr)

        return config

    def requirements_discover(self, direct_requirements):
        direct_requirements = self.translate_requirements(direct_requirements, select_any=False)
        raw_graph = self.requirements_graph(direct_requirements)

        graph = self.first_acceptable_configuration(raw_graph)
      
        return [n for n in graph.nodes() if not graph.node[n]['pruned']], graph


#!/usr/bin/env python

from setuptools import setup

setup(	name='pony-express',
        license = "LGPL-3",
        version='0.2.5.a2',
        description='General purpose package and dependency manager based on MongoDB',
        long_description="""
Artifact PonyExpress (or simply Pony) is an artifact repository and package manager that aims to simplify software modules integration into big projects. It is designed with native (read modern C++) development in mind but it is usefull everywhere there is an environment to setup in order to transform versioned data using any software.

Note: As pony works well with binary components, the transformation function (the tool used to transform data) itself may be a packaged component.

Pony use MongoDB as backend to store your packages and meta-informations so a running local or remote MongoDB database is needed in order to use pony. The pony's MongoDB database is the **pony_store** and the only collection it uses as a repository is named **packages**.

You can perform two main operation from the pony command line:
  
  - **charge** the pony to deliver your artifact to clients.
  - **deliver** dependencies of the working artifact to users.
""",
        author='Giuseppe Puoti',
        author_email='giuseppe.puoti@gmail.com',
        url='https://github.com/gpuoti/Artifact-PonyExpress',

        classifiers=[
            "Development Status :: 3 - Alpha",
            "Topic :: Software Development :: Version Control",
            "Topic :: System :: Archiving :: Packaging",
            "License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)",
        ],
        
        py_modules=[
            'pony', 
            'bag',
            'config_constrainer',
            'dependencies',
            'alternative_set',

            'pony_scons',
            'cli'
            # list them here when you add any other modules!
            ],
            
        
        
        install_requires = [
            'tabulate', 
            'colorama', 
            'networkx',
            'pymongo'
            # add any other dependency package here in order to let pip install them as installation' side effect.
            ],
            
        scripts = ['pony.py']  
)


import sys, os, os.path, collections
import tarfile, json
from fnmatch import fnmatch
import io

import colorama
from tabulate import tabulate

import bag as bag
from bag import Bag, MongoConnectionInfo
import dependencies


silent=False

class MoveOp:
    def __init__(self, source, destination):
        self.source = source
        self.destination = destination
        
    def __eq__(self, other):
        return self.source == other.source
        
    def __hash__(self):
        return self.source.__hash__()
        
def move_ops (instructions, relative_path=''):
    """ensure move operation from package/unpackage instructions are unique.
    Consider that move operations are considered equal if they share the source. That is you can't pack
    the same file into two different places neighter you can unpack the same file in multiple places.
    A move operation from successive instruction in the list will overrid any previous one.
    """
    
    ops = set()
    for i in instructions:
        ops = ops.difference(i.file_op())   # abort any previous discovered move operation on same source
        ops = ops.union(i.file_op())

    return ops

class PackageInstruction:
    def __init__(self, source_folder, filter, destination_folder):
        self.source_folder = os.path.normpath(source_folder)
        self.destination_folder = os.path.normpath(destination_folder)
        self.filter = filter
    

    def accept_file(self, file_name):
        """Use the filter regular exception to check if the received filename is ineresting or not"""
        file_name = os.path.basename(file_name)
        return fnmatch (file_name, self.filter)
    
    def file_op(self, relative_path=""):
        """
        This function recursivelly create and yield MoveOp to command a move from a source folder to a destination one given a filter on files to consider.
        
        yield MoveOp corresponding to a move operation of files. 
        Source path is absolute and obtained chaining
            source_path + relative_path
        where relative_path is the path traversed untill now by this recursive function
        Destination path is obtained chaining 
            destination_path + relative_path
            
        """
        ops = []
        path = os.path.join(self.source_folder , relative_path)
        # a first filter on file from current folder to get candidate to be packed.
        acceptable_generic_files =  [f for f in os.listdir(path) if self.accept_file(f)]
        
        # for each regular file, move and append on the package
        for file_name in [real_file for real_file in acceptable_generic_files if not os.path.isdir(os.path.join(path,real_file))]:
            accepted_file = (os.path.join(relative_path , file_name))
            ops.append(MoveOp(   os.path.join(self.source_folder, accepted_file), 
                                 os.path.join(self.destination_folder, accepted_file))
                       )
            
        # recursivelly process inner folders except special files of course
        folders = [folder for folder in os.listdir(path)  if os.path.isdir(os.path.join(path,folder)) and folder not in ['.','..']  ]

        for folder_name in folders:
            ops += self.file_op(os.path.join(relative_path, folder_name))
            
        return ops
        

def pack(source, destination, filter = '*'):
    """sugar to build MoveOp using a verb"""
    return PackageInstruction(source, filter, destination)
    
unpack = pack


        
    

def gz_package(package_instructions, shelf=None, mem=None):
    """
    Builds a tar.gz package using the received instructions at path indicated by shelf.
    In the proposed methafor, place the package on the shelf following the received instructions.
    """
    
    package = tarfile.open ( shelf, 'w:bz2', fileobj=mem)
        
    repath_log = []
    for op in move_ops(package_instructions):
        package.add( op.source, op.destination)
        repath_log.append( (op.source, op.destination) )
    package.close()
    
    # print the operation log
    if not silent:
        print (tabulate(repath_log, headers= ['', 'saved in']))
    
def gz_unbox(unbox_instructions, package_path=None, mem=None):
    """
    Unpack files from the received package 
    and move them according to the unboxing instructions received
    """
    
    has_common_prefix = os.path.commonprefix
    unbox_log = []
    with tarfile.open(package_path, 'r:bz2', mem) as package:
        for instruction in unbox_instructions:
            source = instruction.source_folder
            for m in [member for member in package.getmembers() if has_common_prefix([source, member.path]) == source]:
                new_path = instruction.destination_folder + m.path[len(source):]
                unbox_log.append ( (m.path, new_path) )
                m.path = new_path
        
        
        package.extractall()
        
    # print the operation log
    print (tabulate(unbox_log, headers= ['', 'unboxed in']))
    return [move[1] for move in unbox_log]
    
def load_metadata(file_path='meta.json') :
    """Load metadata from a json file on disk in the form of a dictionary."""
    with open(file_path) as data_file:    
        meta = json.load(data_file)
    
    return meta
     
def charge(instructions=None, metainfo=None, json=None, file =None, db_connection_info=MongoConnectionInfo()):
    if(json):
        charge_json(json, db_connection_info=db_connection_info)
    elif(instructions and metainfo):
        charge_low_level(instructions, metainfo, db_connection_info=db_connection_info )
    elif(file):
        charge_json_file(file, db_connection_info=db_connection_info)
   
def charge_low_level(instructions, metainfo, db_connection_info=MongoConnectionInfo()):
    package_content = io.BytesIO()
    connected_bag = Bag(db_connection_info)
    if connected_bag.check(metainfo):
        raise bag.YetInBag(metainfo)
    gz_package(package_instructions = instructions, mem = package_content)
    connected_bag.charge(package_content.getvalue(), metainfo)
    
    if not silent:
        print ("package was " + str(int(len(package_content.getvalue()) / 1024)) + "kB") 

def charge_json_file(path, db_connection_info=MongoConnectionInfo()):
    f = open(path, "r")
    charge_json(f.read(), db_connection_info=db_connection_info)
    f.close()
    
    
def charge_json(json_text, db_connection_info=MongoConnectionInfo()):
    jobject = json.loads(json_text)
    box_instructions_json = jobject["__BOX_INSTRUCTIONS__"]
    instructions = []
    for json_instruction in box_instructions_json:
        instructions.append( pack(json_instruction['FROM'], json_instruction['TO'], json_instruction['FILTER']))
    try:
        del jobject["__BOX_INSTRUCTIONS__"]
        del jobject["__UNBOX_INSTRUCTIONS__"]
    except KeyError:
        pass
        
    charge(instructions, jobject, db_connection_info=db_connection_info)
    
def deliver_one(instructions, meta_request, db_connection_info=MongoConnectionInfo()):
    
    connected_bag = Bag(db_connection_info)
    package = connected_bag.take(meta_request)
    package_content = io.BytesIO(package)
    return gz_unbox(unbox_instructions = instructions, mem = package_content)
    
def deliver(instructions, meta_requests, db_connection_info=MongoConnectionInfo()):
    """deliver artifacts defined by each of meta_requests then apply unbox instructions to each."""
    created_files = []
    connected_bag = Bag(db_connection_info)
    
    if isinstance(meta_requests, dict):
        print ('metarequest is a dict, let make a list!')
        meta_requests = [meta_requests]

    direct_requirements = [dependencies.Requirement(r) for r in meta_requests]
    meta_request = [dr.meta_request() for dr in direct_requirements]
    meta_requests, gr =connected_bag.requirements_discover( direct_requirements)
    print(
        'results from bag: ' + str(meta_requests) + '   ' + str(gr)
    )
    
    if isinstance(meta_requests, list):
        for r in meta_requests:
            print("REQUIRE> "+ str(r))
            created_files += deliver_one(instructions, r, db_connection_info=db_connection_info)
    else:
        print("REQUIRE FAILS> "+ str(meta_requests))
    
    return [s.replace('\\','/') for s in created_files], gr

def deliver_json(json_text, db_connection_info=MongoConnectionInfo()):
    jobject = json.loads(json_text)
    unbox_instructions_json = jobject["__UNBOX_INSTRUCTIONS__"]
    instructions = []
    for json_instruction in unbox_instructions_json:
        instructions.append( pack(json_instruction['FROM'], json_instruction['TO'], json_instruction['FILTER']))
    try:
        del jobject["__BOX_INSTRUCTIONS__"]
        del jobject["__UNBOX_INSTRUCTIONS__"]
    except KeyError:
        pass
    try:    
        files, gr = deliver(instructions, jobject["DEPENDENCIES"], db_connection_info=db_connection_info)
        print ("deliver log" )
        for f in files:
            print (str(f))
    except KeyError:
        print('no dependencies detected, nothing to deliver')
        return bag.no_dependencies()
    
    print()
    print("dependencies graph")
    print (bag.to_dot_string(gr))

    return   files, gr
            


if __name__ == '__main__':
    import time
    import cli

    start_time = time.time()
    
    arguments = cli.parser.parse_args(sys.argv[1:])
    print(dir(arguments))
    arguments.func(arguments)
    
    end_time = time.time()
    print( """Executed in """ + str(int( (end_time - start_time)) ) + " s")




import networkx as nx
import dependencies as dep

class ConfigurationImpossible(Exception):
    """Exception raised in case pony can't find an acceptable dependency configuration."""

    def __str__ (self):
        return """Can't fulfill all dependencies in a compatible way.
"""

class config_generator:
    def __init__(self, gr):
        self.gr = gr
        self.clusters = dict()

        for n in self.gr.nodes():
            try:
                self.clusters[n.NAME] += [n]
            except KeyError:
                self.clusters[n.NAME] = [n]

            self.gr.node[n]['ndep'] = self._ndep(self.gr.neighbors(n))
        
    
    def _ndep(self,requirements):
        """
        Given a requirement list, calculate the number of projects involved as dependecy.
        """

        dependant_projects = set()
        for r in requirements:
            try:
                dependant_projects.add( r.NAME )
            except KeyError:
                # ignore empty graph node for now
                pass

        return len(dependant_projects)
    def _reduce_graph(gr, removed_alternatives):
        reduced_gr = gr.copy()
        reduced_gr.remove_nodes_from( removed_alternatives)
        return resuced_gr

    def _local_constraint(self, n, configuration):
        assert type(configuration) == dict
        
        local_constrained_configuration = dict()
        for valid_neighbor in  [x for x in self.gr.neighbors(n) if x in configuration[x.NAME] ]:
            try:
                local_constrained_configuration[valid_neighbor.NAME] += [valid_neighbor]
            except KeyError:
                local_constrained_configuration[valid_neighbor.NAME] = [valid_neighbor]

        if len(local_constrained_configuration) < self.gr.node[n]['ndep']:
            raise ConfigurationImpossible

        for variable in local_constrained_configuration:
            configuration[variable] = local_constrained_configuration[variable]
        
        return configuration

    def _constraint_next(self, variables, unconstrained_configuration):
        assert type(unconstrained_configuration) == dict
        
        if len (variables) > 0:
            cluster = variables[0]

            alternatives = unconstrained_configuration[cluster]
            alternatives.sort(reverse = True)
            for alternative in alternatives:
                try:
                    constrained = unconstrained_configuration
                    constrained[cluster] = [alternative]

                    constrained= self._local_constraint(alternative, constrained)
                    for sub_configuration in self._constraint_next(variables[1:], constrained): 
                        yield (alternative,) + sub_configuration

                except ConfigurationImpossible:
                    pass
        else:
            yield ()
                
    def constraint(self):
        alternative_count = 0
        for config in self._constraint_next(list(self.clusters.keys()), self.clusters):
            print (config)
            alternative_count += 1


            gr_config = self.gr.copy()
            for n in [n for n in gr_config.nodes() if n not in config]:
                gr_config.node[n]['pruned'] = True
                
            yield gr_config
            
        if alternative_count == 0:
            raise ConfigurationImpossible